/*
 * $Id: ups.c,v 1.14 2000/08/30 07:58:11 taca Exp $
 *
f * Copyright (C) 2000, JEPRO Co. Ltd.  All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of JEPRO Co., Ltd. nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <sys/cdefs.h>
#ifndef lint
__COPYRIGHT("Copyright (C) 2000, JEPRO Co. Ltd.  All right reserved.");
__RCSID("$Id: ups.c,v 1.14 2000/08/30 07:58:11 taca Exp $");
#endif
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#ifdef HAVE_UUCPLOCK
#ifdef HAVE_LIBUTIL
#include <libutil.h>
#else
#include <util.h>
#endif /* HAVE_LIBUTLI */
#else /* !HAVE_UUCPLOCK */
#include "uucplock.h"
#endif /* HAVE_UUCPLOCK */
#ifdef HAVE_UPS_CONTROL
#include <sys/com_ups.h>
#endif
#include "bits.h"
#include "ups.h"

#ifndef	DEFAULT_INTERVAL
#define	DEFAULT_INTERVAL	1
#endif

static BITS cget_bits __P((char *, char *));
static int set_bits __P((int, BITS));
static int test_bits __P((BITS, int));
static int get_status __P((UPS));
static char *change_bits __P((int prev, int cur));

#ifdef DEBUG
int ups_setbits_debug = 0;
#endif

UPS
ups_new(name, device, timeout, grace)
	char *name;
	char *device;
	int timeout;
	int grace;
{
	static char *file[] = { UPSTAB,
#ifdef DEBUG
				"upstab",
#endif
				NULL, };
	char *ups_cap;
	UPS ups;
	char *s;

	if (cgetent(&ups_cap, file, name) != 0) {
		perror("ups_new: can't read upstab");
		return NULL;
	}

	ups = calloc(sizeof *ups, 1);
	if (ups == NULL) {
		perror("ups_new");
		return NULL;
	}

	ups->name = strdup(name);
	ups->device = device;
	s = strrchr(device, '/');
	ups->lock = (s == NULL)? device: (s + 1);

	ups->status = NONE;
	ups->interval = DEFAULT_INTERVAL;
	ups->fd = -1;
	ups->tty_status = -1;
	ups->fail_timeout = timeout;
	ups->grace_time = grace;

	ups->init_bits = cget_bits(ups_cap, "init");
	ups->startup_bits = cget_bits(ups_cap, "startup");
	ups->status_bits = cget_bits(ups_cap, "status");
	ups->line_bits = cget_bits(ups_cap, "line");
	ups->low_bits = cget_bits(ups_cap, "low");
	ups->sleep_bits = cget_bits(ups_cap, "sleep");

	if (cgetcap(ups_cap, "noaction", ':') != 0) 
		ups->flag |= UPS_NOACTION;
#ifdef HAVE_UPS_CONTROL
	if (cgetcap(ups_cap, "kernel", ':') != 0) 
		ups->flag |= UPS_KERNEL;
#endif
	free(ups_cap);
	return ups;
}

int
ups_open(ups, mode)
	UPS ups;
	int mode;
{
	int r, fd, on;

	if (ups->fd >= 0) {
		syslog(LOG_ERR, "ups_open: ups_open called twice\n");
		return -1;
	}
	r = uu_lock(ups->lock);
	if (r != 0) {
		if (r > 0)
			syslog(LOG_ERR, "ups_open: Device is busy");
		else
			syslog(LOG_ERR, "ups_open: error on lock: %m");
		return -1;
	}

	fd = open(ups->device, mode);
	if (fd < 0) {
		syslog(LOG_ERR, "ups_open: open %m");
		return -1;
	}

	on = 1;
	if (fcntl(fd, F_SETFD, on) < 0) {
		close(fd);
		syslog(LOG_ERR, "ups_open: fcntl %m");
		return -1;
	}
	ups->fd = fd;
	ups->status = INITIAL;
	setproctitle("opened");
	syslog(LOG_DEBUG, "ups_open: UPS is opened");
	return 0;
}

int
ups_init(ups)
	UPS ups;
{
	BITS b;
	int value, initial, result;

	value = get_status(ups);
	if (value < 0)
		return -1;
	syslog(LOG_INFO, "ups_init: status = %03o(%s)", value, bit_to_modem(value));
	
	if (set_bits(ups->fd, ups->init_bits) < 0)
		return -1;

	initial = 1;
	b = ups->startup_bits;
	if (b != NULL) {
		while ((value = get_status(ups)) >= 0) {
			result = test_bits(b, value);
			if (result)
				break;
			else {
				if (initial) {
					syslog(LOG_INFO,
					       "ups_init: status = %03o(%s)",
					       value, bit_to_modem(value));
					syslog(LOG_NOTICE,
					       "ups_init: waiting for UPS ready");
					initial = 0;
					setproctitle("waiting for ready");
				}
				mysleep(ups->interval);
			}
		}
		if (value < 0) {
			syslog(LOG_ERR, "ups_init: %m");
			return -1;
		}
	}
	ups->status = READY;
	setproctitle("ready");
	syslog(LOG_INFO, "ups_init: UPS is ready");
	return 0;
}

int
ups_status(ups)
	UPS ups;
{
	BITS line, low;
	int fd, value, changed;
	char *s;

	line = ups->line_bits;
	low = ups->low_bits;
	fd = ups->fd;

	mysleep(ups->interval);
	value = get_status(ups);
	time(&ups->when);
	if (value < 0)
		goto error;

	if (ups->tty_status == -1) {
		syslog(LOG_INFO, "ups_status: %03o(%s)", value,
		       bit_to_modem(value));
		changed = 1;
	} else if (ups->tty_status != value) {
		s = change_bits(ups->tty_status, value);
		syslog(LOG_INFO, "status changed: %s", s);
		changed = 1;
	} else
		changed = 0;

	if (changed || ups->status == LINE_LOST) {
		if (test_bits(line, value)) {
			if (ups->status == LINE_LOST && test_bits(low, value)) {
				ups->status = BATTERY_LOW;
				setproctitle("low battery");
			} else {
				ups->status = LINE_LOST;
				setproctitle("line power lost");
			}
		} else {
			ups->status = READY;
			setproctitle("ready");
		}
		ups->tty_status = value;
	}
	return 0;
 error:
	ups->status = NONE;
	setproctitle("error");
	return -1;
}

int
ups_sleep(ups)
	UPS ups;
{
	BITS b;
	int status, value;

	if (ups == NULL || ups->fd < 0)
		return 0;
	b = ups->sleep_bits;
	if (ups->flag & UPS_NOACTION) {
		syslog(LOG_INFO, "ups_sleep: no action");
		return 0;
	}

#ifdef HAVE_UPS_CONTROL
	if (ups->flag & UPS_KERNEL) {
		int ovalue;

		value = get_status(ups);
		if (value < 0)
			return -1;
		ovalue = value;
		while (b != NULL) {
			value = bits_data(b, value);
			b = bits_next(b);
		}
		syslog(LOG_INFO, "ups_sleep: set to kernel %03o(%s)", value,
		       bit_to_modem(value));
		if (value != ovalue) {
			if (ioctl(ups->fd, UPSIOC_SET, &value) < 0) {
				syslog(LOG_DEBUG, "UPSIOC_SET %03o(%s): %m",
				       value, bit_to_modem(value));
				return -1;
			}
		}
		return 0;
	}
#endif
	sync(); sync();
	syslog(LOG_DEBUG, "ups_sleep: UPS will sleep");
#ifdef DEBUG
	ups_setbits_debug = 1;
#endif
	status = set_bits(ups->fd, b);
	if (status < 0)
		syslog(LOG_ERR, "ups_sleep: failed to sleep");
	value = get_status(ups);
	if (value < 0)
		return -1;
	syslog(LOG_DEBUG, "ups_sleep: status is now %03o(%s)",
	       value, bit_to_modem(value));
	return status;
}

void
ups_close(ups, keep)
	UPS ups;
	int keep;
{
	if (!keep) {
		(void)uu_unlock(ups->lock);
		if (ups->fd >= 0) {
			(void)close(ups->fd);
			ups->fd = -1;
		}
	}
	free(ups->name);
	free(ups);
	syslog(LOG_DEBUG, "ups_close: finished");
}

static BITS
cget_bits(cap, str)
	char *cap;
	char *str;
{
	BITS b;
	char *s;

	if (cgetstr(cap, str, &s) < 0)
		return NULL;
	b = bits_parse(s);
	free(s);
	return b;
}

static int
set_bits(fd, b)
	int fd;
	BITS b;
{
	int op, result, value;
	unsigned long request;

	result = 0;
	while (result == 0 && b != NULL) {
		request = 0;
		op = bits_op(b);
		value = bits_bits(b);
		switch (op) {
		case BITS_RESET:
			request = TIOCMBIC;
			break;
		case BITS_SET:
			request = TIOCMBIS;
			break;
		default:
			syslog(LOG_ERR,
			       "unknown modem bit operation %d", op);
			break;
		}
#ifdef DEBUG
		if (ups_setbits_debug)
		    syslog(LOG_DEBUG, "%s %03o(%s)",
			   (request == TIOCMBIC)? "clear":
			   (request == TIOCMBIS)? "set": "what?",
			   value, bit_to_modem(value));
#endif
		result = ioctl(fd, request, &value);
		if (result < 0) {
		    syslog(LOG_ERR, "ioctl set (%lx, %o): %m", request, value);
		}
		b = bits_next(b);
	}
	return result;
}

static int
get_status(ups)
	UPS ups;
{
	int fd, result, value;
	BITS s;

	fd = ups->fd;
	s = ups->status_bits;
	if (set_bits(fd, s) < 0)
		return -1;

	result = ioctl(fd, TIOCMGET, &value);
	if (result < 0) {
		syslog(LOG_ERR, "ioctl get (TIOCMGET): %m");
		return result;
	}
	return value;
}

static int
test_bits(b, value)
	BITS b;
	int value;
{
	int result;

	result = 0;
	while (b != NULL) {
		result = bits_test(b, value);
		if (!result)
			break;
		b = bits_next(b);
	}
	return result;
}

void
mysleep(second)
	unsigned int second;
{
	struct timeval timeout;

	timeout.tv_sec = second;
	timeout.tv_usec = 0;

	if (select(0, NULL, NULL, NULL, &timeout) < 0) {
		if (errno != EINTR) {
			syslog(LOG_ERR, "select: %m");
		}
	}
}

static char *
change_bits(prev, cur)
	int prev;
	int cur;
{
	static char buf[30];
	char *s, *u;
	int set, reset, m, n;

	set = ((prev ^ cur) & cur);
	reset = ((prev ^ cur) & prev);

	*buf = '\0';
	n = sizeof buf - 1;
	u = buf;
	if (set) {
		s = bit_to_modem(set);
		if (strchr(s, '|') == NULL) {
			snprintf(u, n, "+%03o(%s)", set, s);
		} else {
			snprintf(u, n, "+%03o(%s)", set, s);
		}
		m = strlen(u);
		n -= m;
		u += m;
	}
	if (reset) {
		if (set) {
			strcat(u, ",");
			u++;
			n--;
		}
		s = bit_to_modem(reset);
		if (strchr(s, '|') == NULL) {
			snprintf(u, n, "-%03o(%s)", reset, s);
		} else {
			snprintf(u, n, "-%03o(%s)", reset, s);
		}
	}
	return buf;
}
